import numpy as np
import scipy.stats as stat


class Stat:

    @staticmethod
    def mean_value(s):
        return np.mean(s)

    @staticmethod
    def std_value(s):
        return np.std(s)

    @staticmethod
    def rms_value(s):
        return np.sqrt(np.mean(s**2))

    @staticmethod
    def skew_value(s):
        return stat.skew(s)

    @staticmethod
    def kurtosis_value(s):
        return stat.kurtosis(s)

    @staticmethod
    def crest_value(s):
        return np.abs(s).max()/np.sqrt(np.mean(s**2))

    @staticmethod
    def clearance_value(s):
        return np.abs(s).max()/np.mean(np.sqrt(np.abs(s)))**2

    @staticmethod
    def shape_value(s):
        return np.sqrt(np.mean(s**2))/np.mean(np.abs(s))

    @staticmethod
    def impulse_value(s):
        return np.abs(s).max()/np.mean(np.abs(s))


class SignalCalculator(Stat):

    def __init__(self, data: np.ndarray, **kwargs):
        """ Calculate the sliding value of a signal
            with the window and step

        Args:
            data (np.ndarray): 1d
            window (int, optional): default =1
            step (int, optional): default =1
        """
        window = kwargs.get('window', 1)
        step = kwargs.get('step', 1)
        # sliding data
        self.data = self.sliding(data, window, step)

    @staticmethod
    def sliding(data, window, step):
        return np.lib.stride_tricks.sliding_window_view(data, window)[::step, :]

    def mean_sliding(self):
        return np.mean(self.data, axis=-1)  # axis=-1 means the last dimension

    def std_sliding(self):
        return np.std(self.data, axis=-1)

    def rms_sliding(self):
        return np.apply_along_axis(self.rms_value, -1, self.data)

    def skew_sliding(self):
        return np.apply_along_axis(self.skew_value, -1, self.data)

    def kurtosis_sliding(self):
        return np.apply_along_axis(self.kurtosis_value, -1, self.data)

    def crest_sliding(self):
        return np.apply_along_axis(self.crest_value, -1, self.data)

    def clearance_sliding(self):
        return np.apply_along_axis(self.clearance_value, -1, self.data)

    def shape_sliding(self):
        return np.apply_along_axis(self.shape_value, -1, self.data)

    def impulse_sliding(self):
        return np.apply_along_axis(self.impulse_value, -1, self.data)



import numpy as np
# from scipy import signal
from scipy.fft import rfft, rfftfreq
from scipy.signal import spectrogram, butter, lfilter, freqz


class FrequencyAnalysis:

    @staticmethod
    def calc_fft(data, sample_rate, norm=None):
        """Calculate FFT. Transfrom Signal data from
        Time domain to Frequenct domain.

        Args:
            data (numpy.ndarray): Signal data in Time domain
            sample_rate (scalar): Sampling rate [Hz]

        Returns:
            1. numpy.ndarray: x-axis value: frequency
            2. numpy.ndarray: y-axis value: magnitude
        """
        sample_size = len(data)

        # yf = fft(data)
        # xf = fftfreq(sample_size, 1 / sample_rate)

        yf = rfft(data, norm=norm)  # only get the right side
        # yf = rfft(data, norm='forward')  # norm='forward' 'ortho'

        # as used rfft, need to use to rfftfreq to map
        xf = rfftfreq(sample_size, 1 / sample_rate)

        # How to plot:
        # plt.plot(xf, np.abs(yf))
        # plt.show()
        return xf, np.abs(yf)

    @staticmethod
    def calc_spectogram(data, sample_rate):
        f, t, Sxx = spectrogram(data, sample_rate)

        # How to plot:
        # fig, ax = plt.subplots()
        # spectro = ax.pcolormesh(t, f, Sxx, shading='gouraud')
        # fig.colorbar(spectro, label='|FFT Amplitude|')
        # ax.set_ylabel('Frequency [Hz]')
        # ax.set_xlabel('Time [sec]')
        # plt.show()
        return f, t, Sxx

    # == Bandpass Filter == #

    @staticmethod
    def bandpass_butter(lowcut, highcut, fs, order=5):
        return butter(order, [lowcut, highcut], fs=fs, btype='band')

    @staticmethod
    def bandpass_filter(data, lowcut, highcut, fs, order=5):
        b, a = FrequencyAnalysis.bandpass_butter(
            lowcut, highcut, fs, order=order)
        y = lfilter(b, a, data)
        return y

    # == Lowpass Filter == #

    @staticmethod
    def lowpass_butter(cutoff, fs, order=5):
        return butter(order, cutoff, fs=fs, btype='low', analog=False)

    @staticmethod
    def lowpass_filter(data, cutoff, fs, order=5):
        b, a = FrequencyAnalysis.lowpass_butter(cutoff, fs, order=order)
        y = lfilter(b, a, data)
        return y

    # == Visualize the Bandpass Filter == #

    @staticmethod
    def plot_bandpass_butter(lowcut, highcut, fs, order, worN=8000):

        b, a = FrequencyAnalysis.bandpass_butter(lowcut, highcut, fs, order)
        w, h = freqz(b, a, fs=fs, worN=worN)

        # How to plot:
        # plt.plot(w, abs(h))
        # plt.show()
        return w, abs(h)

    # == Visualize the Lowpass Filter == #

    @staticmethod
    def plot_lowpass_butter(cutoff, fs, order, worN=8000):

        b, a = FrequencyAnalysis.lowpass_butter(cutoff, fs, order)
        w, h = freqz(b, a, fs=fs, worN=worN)

        # How to plot:
        # plt.plot(w, abs(h))
        # plt.show()
        return w, abs(h)
